import os
import subprocess
from operator import itemgetter


class Fio(object):
    def __init__(self, data_dir, filesize='100m', iodepth=12, runtime='1h', read_rate='1g', write_rate='1g',
                 initialize=True):
        """Initialize the FIO wrapper class

        :param data_dir: Where the test data file goes, which indicates what disk to test
        :param filesize: Size of the test data file, in fio notation format
        :param iodepth: How many blocks to keep in flight during the test
        :param runtime: How long to run the test
        :param read_rate: Static value to set reads at
        :param write_rate: Static value to set writes at
        :param initialize: Bool choose not to initialize the fio data file at class creation
        """
        self.directory = data_dir
        self.filesize = filesize
        self.iodepth = iodepth
        self.runtime = runtime
        self.read_rate = read_rate
        self.write_rate = write_rate
        self.configfile = os.path.dirname(os.path.abspath(__file__)) + '/resources/config.fio'
        if initialize:
            self._execute(init=True)

    def _execute(self, init=False, read_rate=None, write_rate=None, read_type='random', write_type='sequential'):
        """ Private function to execute fio

        :param init: Only init the data file and exit
        :param read_rate: Static value to set reads at
        :param write_rate:  static value to set writes at
        :param read_type: random or sequential
        :param write_type:  random or sequential
        :return: stdout of fio
        """
        if read_type == 'random':
            rt = 'randread'
        elif read_type == 'sequential':
            rt = 'read'
        else:
            raise Exception("Invalid Read Type")
        if write_type == 'random':
            wt = 'randwrite'
        elif write_type == 'sequential':
            wt = 'write'
        else:
            raise Exception("Invalid Write Type")

        command = "/usr/bin/fio"
        options = " --minimal" \
                  " --directory=" + str(self.directory) + \
                  " --filesize=" + str(self.filesize) + \
                  " --iodepth=" + str(self.iodepth) + \
                  " --time_based" \
                  " --runtime=" + str(self.runtime) + \
                  " --ioengine=libaio" \
                  " --blocksize_range=4k-26k,40k-40k" \
                  " --filename=cbrqualifier.0" \
                  " --norandommap" \
                  " --direct=1" \
                  " --randrepeat=0" \
                  " --size=20t" \
                  " --name=read" \
                  " --rw=" + str(rt) + \
                  " --name=write" \
                  " --rw=" + str(wt)
        if not read_rate:
            read_rate = self.read_rate
        if not write_rate:
            write_rate = self.write_rate
        options = " --rate=" + str(read_rate) + "," + str(write_rate) + \
                  options
        if init:
            options = " --create_only=1 " + \
                      options
        proc = subprocess.Popen(command + options, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        if proc.returncode > 0:
            raise Exception("Failed to execute fio correctly\n fio Return Code: %d\nSTDERR from fio:\n %s" %
                            (proc.returncode, stderr))
        return stdout

    @staticmethod
    def _parse_fio_output(output):
        """ Parse output of fio

        :param output: stdout from fio
        :return: tuple with Read Bandwidth, Read IOPS, Write Bandwidth, and Write IOPS from the test
        """
        ret_read_bandwidth, ret_read_iops, ret_write_bandwidth, ret_write_iops = None, None, None, None
        for line in output.split("\n")[0:2]:
            jobname, read_bandwidth, read_iops, write_bandwidth, write_iops = \
                itemgetter(2, 6, 7, 47, 48)(line.split(";"))
            if jobname == "read":
                ret_read_bandwidth = read_bandwidth
                ret_read_iops = read_iops
            elif jobname == "write":
                ret_write_bandwidth = write_bandwidth
                ret_write_iops = write_iops
        return int(ret_read_bandwidth) * 1024, int(ret_read_iops), \
            int(ret_write_bandwidth) * 1024, int(ret_write_iops)

    def run(self, read_rate=None, write_rate=None, write_type='sequential', read_type='random'):
        """

        :param read_rate: Static value to set reads at
        :param write_rate:  static value to set writes at
        :param read_type: random or sequential
        :param write_type:  random or sequential
        :return: tuple with Read Bandwidth, Read IOPS, Write Bandwidth, and Write IOPS from the test
        """
        return self._parse_fio_output(self._execute(read_rate=read_rate, write_rate=write_rate,
                                                    write_type=write_type, read_type=read_type))

    def delete_testfile(self):
        """Delete the test datafile """
        os.unlink(self.directory + '/cbrqualifier.0')
