import os
import yaml
from cbr_qualifier.disk.fio_wrapper import Fio


def dehumanize(hvalue):
    """ Take fio style human specifiers for rate and convert to bytes

    :param hvalue: Human style fio value
    :return: Value in bytes
    """
    conversions = {'k': 1024, 'm': 1024**2, 'g': 1024**3, 't': 1024**4}
    try:
        int(hvalue)
    except ValueError:
        value = int(hvalue[:-1])
        unit = hvalue[-1:]
        return value * conversions[unit]
    else:
        return int(hvalue)


class TestRunner(object):

    default_conf_file = os.path.dirname(os.path.abspath(__file__)) + '/resources/configs.yml'
    with open(default_conf_file, 'r') as stream:
        default_conf = yaml.load(stream)
    run_config = '1000_hosts'

    def __init__(self, data_dir, conf_file=None, runtime=None, close_enough_factor=None, initialize=True, profile=None,
                 multiplier=None, filesize=None):
        """Init TestRunner class

        :param data_dir:  Location of fio test file
        :param conf_file: Optional specification of a config file, use config inside package if not specified
        :param runtime: Optional runtime, use runtime in config if not specified
        :param close_enough_factor: Optional fudge factor, use .95 if not specified
        :param initialize: Initialze Datafile
        :param profile: Profile in config to run 1000_hosts profile used if not specified
        :param multiplier: Multipler for rate values in tests, 1 used if not specified
        :param filesize: Filesize of the test datafile, use filesize in config if not specified
        """
        if conf_file:
            self.conf = self.read_config(conf_file)
        else:
            self.conf = TestRunner.default_conf
        if profile:
            self.run_config = profile
        else:
            self.run_config = TestRunner.run_config
        if not runtime:
            self.runtime = self.conf[self.run_config]['testrun_length']
        else:
            self.runtime = runtime
        if not filesize:
            self.filesize = self.conf[self.run_config]['testdata_file_size']
        else:
            self.filesize = filesize
        if not multiplier:
            self.multiplier = 1
        else:
            self.multiplier = int(multiplier)
        print("Running Profie %s" % self.run_config)
        print("Initializing Datafile of %s, for 1t expect 2-3 hours" % self.filesize)
        self.testfio = Fio(data_dir=data_dir, runtime=self.runtime, initialize=initialize,
                           filesize=self.filesize)
        if not close_enough_factor:
            self.close_enough_factor = .95

    @staticmethod
    def read_config(config_file):
        """ Read the yaml config file

        :param config_file:  config file
        :return: Nested dictionary of the config file contents
        """
        with open(config_file, 'r') as stream:
            conf = yaml.load(stream)
        return conf

    def _pass_fail(self, results):
        """ Calculate pass/fail for results

        :param results: results dictionary
        :return: modified with pass/fail results dictionary
        """
        for result in results:
            if results[result]['result'] > results[result]['threshold'] * self.close_enough_factor:
                results[result]['pass'] = True
            else:
                results[result]['pass'] = False
        return results

    def run_test(self, test):
        """ Run a specific test in a profile

        :param test: Test Name to run
        :return: Dictionary containing test results
        """
        print("Running %s test for %s" % (test, self.runtime))
        results = {'read_bandwidth': {}, 'read_iops': {}, 'write_bandwidth': {}, 'write_iops': {}}

        try:
            read_baseline = self.conf[self.run_config]['tests'][test]['read_baseline']
        except KeyError:
            read_baseline = None
        try:
            write_baseline = self.conf[self.run_config]['tests'][test]['write_baseline']
        except KeyError:
            write_baseline = None

        results['read_bandwidth']['result'], \
            results['read_iops']['result'], \
            results['write_bandwidth']['result'], \
            results['write_iops']['result'] \
            = self.testfio.run(write_rate=dehumanize(write_baseline) * self.multiplier,
                               read_rate=dehumanize(read_baseline) * self.multiplier,
                               write_type=self.conf[self.run_config]['tests'][test]['write'],
                               read_type=self.conf[self.run_config]['tests'][test]['read'])

        results['read_bandwidth']['threshold'] = \
            dehumanize(self.conf[self.run_config]['tests'][test]['read_threshold']) * self.multiplier
        results['read_iops']['threshold'] = \
            self.conf[self.run_config]['tests'][test]['read_threshold_iops'] * self.multiplier
        results['write_bandwidth']['threshold'] = \
            dehumanize(self.conf[self.run_config]['tests'][test]['write_threshold']) * self.multiplier
        results['write_iops']['threshold'] = \
            self.conf[self.run_config]['tests'][test]['write_threshold_iops'] * self.multiplier

        results = self._pass_fail(results)

        results['run_config'] = self.conf[self.run_config]['tests'][test]
        results['run_config']['testdata_file_size'] = self.filesize
        results['run_config']['runtime'] = self.runtime
        results['run_config']['multiplier'] = self.multiplier
        return {test: results}

    def get_tests(self):
        """

        :return: Names of tests in the profile
        """
        return self.conf[self.run_config]['tests'].keys()

    def run(self):
        """ Run all tests in the profile

        :return: list of dictionaries with test results
        """
        test_results = []
        for test in sorted(self.get_tests()):
            test_results.append(self.run_test(test))
        return test_results

    def get_run_configs(self):
        """ Get Profiles in config

        :return: List of profiles in config
        """
        return self.conf.keys()

    def set_run_config(self, run_config):
        """ Set the profile to run

        :param run_config:  Profile to run
        :return: None
        """
        if run_config in self.conf.keys():
            self.run_config = run_config
        else:
            raise Exception("Invalid Run Config")

    def delete_testfile(self):
        """Delete the fio testfile"""
        self.testfio.delete_testfile()

    @staticmethod
    def print_tests(conf_file, run_config):
        """Print a list of tests in the config and profile to screen"""
        if not run_config:
            run_config = TestRunner.run_config
        if conf_file:
            tests = TestRunner.read_config(conf_file)[run_config]['tests']
        else:
            tests = TestRunner.default_conf[run_config]['tests']
        template = '{0:25}{1:30}'
        for test in sorted(tests):
            print(template.format(test, tests[test]['description']))

    @staticmethod
    def print_profiles(conf_file):
        """print a list of profiles in the config to screen"""
        if conf_file:
            profiles = TestRunner.read_config(conf_file)
        else:
            profiles = TestRunner.default_conf
        template = '{0:25}{1:30}'
        for profile in sorted(profiles):
            print(template.format(profile, profiles[profile]['description']))

    @staticmethod
    def print_default_config_yaml():
        """print to screen the yaml from the default config in the package"""
        print(yaml.dump(TestRunner.default_conf, default_flow_style=False))

    @staticmethod
    def print_results(results=None, verbose=False, results_file=None):
        """ Print to screen results

        :param results:  list of dictionaries with results
        :param verbose:  Bool, print verbose items
        :param results_file: yaml results output file, this will override results
        :return: None
        """
        if results_file:
            results = TestRunner.read_results(results_file)
        template = "{0:25}{1:>15}{2:>15}{3:>15}{4:>15}"
        for test in sorted(results):
            test_name = test.keys()[0]
            print("Test: %s - %s" % (test_name, test[test_name]['run_config']['description']))
            print("\tRead_type: %s" % test[test_name]['run_config']['read'])
            print("\tWrite_type: %s" % test[test_name]['run_config']['write'])
            try:
                print("\tRead Baseline: %d MB/s" %
                      (dehumanize(test[test_name]['run_config']['read_baseline']) *
                       test[test_name]['run_config']['multiplier']/1024/1024))
            except KeyError:
                print("\tRead Baseline: N/A")
            try:
                print("\tWrite Baseline: %d MB/s" %
                      (dehumanize(test[test_name]['run_config']['write_baseline']) *
                       test[test_name]['run_config']['multiplier']/1024/1024))
            except KeyError:
                print("\tWrite Baseline: N/A")
            print("\tData File Size: %s" % test[test_name]['run_config']['testdata_file_size'])
            print("\tRuntime: %s" % test[test_name]['run_config']['runtime'])
            print("\tMultiplier: %s" % test[test_name]['run_config']['multiplier'])
            print(template.format('Metric', 'Pass/Fail', 'Threshold', 'Test Value', "Percent"))
            test_pass = True
            for metric in sorted(test[test_name]):
                if metric == 'write_bandwidth' or metric == 'read_bandwidth':
                    print(template.format(metric + ' MB/s', str(test[test_name][metric]['pass']),
                                          test[test_name][metric]['threshold']/1024/1024,
                                          test[test_name][metric]['result']/1024/1024,
                                          100*test[test_name][metric]['result']/test[test_name][metric]['threshold']))
                    if not test[test_name][metric]['pass']:
                        test_pass = False
                if verbose:
                    if metric == 'write_iops' or metric == 'read_iops':
                        print(template.format(metric, str(test[test_name][metric]['pass']),
                                              test[test_name][metric]['threshold'],
                                              test[test_name][metric]['result'],
                                              100*test[test_name][metric]['result'] /
                                              test[test_name][metric]['threshold']))
                        if not test[test_name][metric]['pass']:
                            test_pass = False
            if not test_pass:
                print("\n%s Result:     %s" % (test_name, test[test_name]['run_config']['fail_message']))
            else:
                print("\n%s Result:     %s" % (test_name, test[test_name]['run_config']['pass_message']))
            print("\n")

    @staticmethod
    def write_results(results, out_file):
        """Write results to yaml file"""
        with open(out_file, 'w') as f:
            yaml.dump(results, f, default_flow_style=False)

    @staticmethod
    def read_results(in_file):
        """Read results from yaml file"""
        with open(in_file, 'r') as f:
            results = yaml.load(f)
        return results
